package com.fanruan.proxy.interceptor;

import com.esotericsoftware.kryo.Kryo;
import com.fanruan.annotation.*;
import com.fanruan.pojo.message.RpcRequest;
import com.fanruan.service.jdbc.AbstractBind;
import com.fanruan.service.jdbc.ServiceInputStream;
import com.fanruan.service.jdbc.ServiceReader;
import com.fanruan.utils.Commons;

import java.io.Reader;
import java.lang.reflect.Method;
import java.lang.reflect.Parameter;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.Map;

/**
 * @author Yichen Dai
 */
public class InterceptorUtils {

    private static final Map<String, Class<?>> WRAPPER_CLASS_MAP = new HashMap<String, Class<?>>(){
        {
            put("Integer", Integer.TYPE);
            put("Short", Short.TYPE);
            put("Long", Long.TYPE);
            put("Double", Double.TYPE);
            put("Float", Float.TYPE);
            put("Byte", Byte.TYPE);
            put("Character", Character.TYPE);
            put("Boolean", Boolean.TYPE);
        }
    };

    public static final Kryo kryo = new Kryo();

    public static boolean isWraps(Object o){
        Class<?> clz = o.getClass();
        if(clz == null) {
            return false;
        }
        return WRAPPER_CLASS_MAP.containsKey(getClassName(clz.getName()));
    }


    public static String getClassName(String fullyQualifiedClassName){
        String[] arr = fullyQualifiedClassName.split("\\.");
        int n = arr.length;
        if(n == 0) {
            throw new RuntimeException("the class name invoked is wrong");
        }
        return arr[n-1];
    }

    public static boolean isInBindList(Object o){
        if (o == null) {
            return false;
        }
        Class<?> clazz = o.getClass();
        return clazz.isAnnotationPresent(RemoteClass.class);
    }

    public static void replaceBindingParameter(Method method, Object[] args){
        Parameter[] pars = method.getParameters();
        for(int i=0; i<pars.length; i++){
            Parameter par = pars[i];
            if(par.isAnnotationPresent(BindingParamter.class) && isInBindList(args[i])){
                args[i] = getBindID(args[i]);
            }
        }
    }

    public static void setBindID(Object returnObj, String ID){
        if(returnObj instanceof AbstractBind){
            AbstractBind ab = (AbstractBind) returnObj;
            ab.setID(ID);
        }else if(returnObj instanceof ServiceReader){
            ServiceReader sr = (ServiceReader) returnObj;
            sr.setID(ID);
        }else if(returnObj instanceof ServiceInputStream){
            ServiceInputStream sis = (ServiceInputStream) returnObj;
            sis.setID(ID);
        }
    }

    public static String getBindID(Object o){
        String idToInvoke = null;
        if(o instanceof AbstractBind){
            AbstractBind ab = (AbstractBind) o;
            idToInvoke = ab.getID();
        }else if(o instanceof Reader){
            ServiceReader sr = (ServiceReader) o;
            idToInvoke = sr.getID();
        }else if(o instanceof ServiceInputStream){
            ServiceInputStream sis = (ServiceInputStream) o;
            idToInvoke = sis.getID();
        }
        return idToInvoke;
    }

    public static RpcRequest generateRequest(Class<?> clazz, Object o, Method method, Object[] objects) throws SQLException {
        checkObject(method, objects);

        RpcRequest rpcRequest = new RpcRequest();
        rpcRequest
                .setID(Commons.getID())
                .setMethodName(method.getName())
                .setArgs(objects)
                .setArgTypes(method.getParameterTypes());

        // get RPC class name from annotation
        // if null, send the name of this clazz
        if(clazz.isAnnotationPresent(RemoteClass.class)){
            RemoteClass annotation = clazz.getAnnotation(RemoteClass.class);
            rpcRequest.setServiceClassName(annotation.remoteClassName());
        }else{
            rpcRequest.setServiceClassName(clazz.getName());
        }

        // IDtoInvoke is an unique ID to identify an one-to-one binding relation.
        // It comes from rpcRequest in which the instance in the agent is created.
        String idToInvoke = getBindID(o);

        if(idToInvoke != null){
            rpcRequest.setIDToInvoke(idToInvoke);
        }
        return rpcRequest;
    }

    public static void checkObject(Method method, Object[] objects) throws SQLException {
        for (int i=0; i<objects.length; i++){
            Object obj = objects[i];
            if(obj == null){
                continue;
            }
            try{
                if(kryo.getDefaultSerializer(obj.getClass()) != null){
                  continue;
                }
            }catch (IllegalArgumentException e){
                // When kryo hasn't default serializer to handler obj, the IllegalArgumentException will be thrown.
            }

            try {
                obj.getClass().getDeclaredConstructor();
            }catch (NoSuchMethodException e){
                throw new SQLException("can't invoke " + method.getName() + "dose not have no-arg constructor");
            }

        }
    }

    public static boolean isNotImplemented(Method method) {
        return method.isAnnotationPresent(NotImplemented.class);
    }

    public static boolean isLocalMethod(Method method) {
        return method.isAnnotationPresent(LocalMethod.class);
    }

    public static boolean isWithBindingParameter(Method method){ return method.isAnnotationPresent(WithBindingParameter.class); }
}
