package com.fanruan;

import com.fanruan.cache.BeanCacheImpl;
import com.fanruan.exception.NotImplementedException;
import com.fanruan.handler.DispatcherImpl;
import com.fanruan.pojo.message.RpcRequest;
import com.fanruan.service.jdbc.connection.ServiceConnection;
import org.junit.jupiter.api.*;

import java.sql.*;
import java.util.Collections;
import java.util.HashMap;
import java.util.Properties;

/**
 * @author Yichen Dai
 * @date 2022/8/30 10:23
 */
@TestInstance(TestInstance.Lifecycle.PER_CLASS)
@TestMethodOrder(MethodOrderer.OrderAnnotation.class)
public class ConnectionTest extends BaseJDBCTest{

    Connection connection = null;
    String IDtoInvoke = "";
    BeanCacheImpl beanCache = null;
    DispatcherImpl dispatcher = null;
    ServiceConnection serviceConnection = null;

    @BeforeAll
    public void setUp() throws SQLException {
        openSocket();
        connection = getConnection();
        ServiceConnection conn = (ServiceConnection) connection;
        IDtoInvoke = conn.getID();
        dispatcher = agent.dispatcherImpl;
        beanCache = dispatcher.CACHE.getBeanCache(dbName);
        serviceConnection = (ServiceConnection) ProxyFactoryIT.getProxy(ServiceConnection.class, map);
    }

    @Test
    public void testSetCatalogSchema() throws Throwable{
        String db = connection.getCatalog();
        String schema = connection.getSchema();
        connection.setCatalog(db);
        connection.setSchema("PUBLIC");

        // get the current schema
        ResultSet rst = connection.createStatement().executeQuery("SELECT * " +
                "FROM   INFORMATION_SCHEMA.TABLES");
        Assertions.assertTrue(rst.next());
        Assertions.assertEquals("PUBLIC", rst.getString(1));
        Assertions.assertEquals(db, connection.getCatalog());
        Assertions.assertEquals("PUBLIC", connection.getSchema());

        // get the current schema
        connection.setSchema(schema);
        rst = connection.createStatement().executeQuery("SELECT * " +
                "FROM   INFORMATION_SCHEMA.TABLES");
        Assertions.assertTrue(rst.next());
        Assertions.assertEquals(schema, rst.getString(1));
        rst.close();
    }

    @Test
    public void testConnection() throws Throwable {
        Properties property = connection.getClientInfo();
        Assertions.assertNull(property);
        Properties clientInfo = new Properties();
        clientInfo.setProperty("name", "Peter");
        clientInfo.setProperty("description", "HSQLDB JDBC");

        Assertions.assertThrows(SQLClientInfoException.class, () -> {
            serviceConnection.setClientInfo("name", "peter");
            RpcRequest request = map.get(null);
            request.setIDToInvoke(IDtoInvoke);
            dispatcher.invokeAsRequest(request, beanCache);
        });

        Assertions.assertThrows(SQLClientInfoException.class, () -> {
            serviceConnection.setClientInfo(clientInfo);
            RpcRequest request = map.get(null);
            request.setIDToInvoke(IDtoInvoke);
            dispatcher.invokeAsRequest(request, beanCache);
        });
    }

    @Test
    public void testNetworkTimeout() throws SQLException {
        int millis = connection.getNetworkTimeout();
        Assertions.assertEquals(0, millis);
        Assertions.assertThrows(SQLFeatureNotSupportedException.class, () -> {
            serviceConnection.setNetworkTimeout(null, 200);
            RpcRequest request = map.get(null);
            request.setIDToInvoke(IDtoInvoke);
            dispatcher.invokeAsRequest(request, beanCache);
        });
    }

    @Test
    public void testAbort() throws SQLException {
        Assertions.assertThrows(SQLException.class, () -> {
            serviceConnection.abort(null);
            RpcRequest request = map.get(null);
            request.setIDToInvoke(IDtoInvoke);
            dispatcher.invokeAsRequest(request, beanCache);
        });
    }

    @Test
    public void testClose() throws SQLException {
        Connection conn = getConnection();
        Assertions.assertFalse(conn.isClosed());
        conn.close();
        Assertions.assertTrue(conn.isClosed());
    }

    @Test
    public void testHoldability() throws Throwable {
        connection.setHoldability(ResultSet.CLOSE_CURSORS_AT_COMMIT);
        Assertions.assertEquals(ResultSet.CLOSE_CURSORS_AT_COMMIT, connection.getHoldability());
    }

    @Test
    public void testIsValid() throws Throwable {
        Assertions.assertTrue(connection.isValid(10));
    }

    @Test
    public void testUnwrapper() throws SQLException {
        boolean canUnwrap = connection.isWrapperFor(Connection.class);
        Assertions.assertTrue(canUnwrap);
        Assertions.assertThrows(SQLException.class,
                () -> connection.unwrap(null));
    }

    @Test
    public void testNativeSQL() throws SQLException {
        Assertions.assertEquals("select 1", connection.nativeSQL("select 1"));
    }

    @Test
    public void testTypeMap() throws Throwable {
        Assertions.assertEquals(Collections.emptyMap(), connection.getTypeMap());
        Assertions.assertThrows(SQLException.class, () -> {
            serviceConnection.setTypeMap(new HashMap<>());
            RpcRequest request = map.get(null);
            request.setIDToInvoke(IDtoInvoke);
            dispatcher.invokeAsRequest(request, beanCache);
        });
    }

    @Test
    public void setReadOnly() throws SQLException {
        Assertions.assertFalse(connection.isReadOnly());
        connection.setReadOnly(true);
        Assertions.assertTrue(connection.isReadOnly());
    }

    @Test
    public void setTransactionIsolation() throws SQLException {
        Connection conn = getConnection();
        Assertions.assertEquals(Connection.TRANSACTION_READ_COMMITTED ,conn.getTransactionIsolation());
        conn.setTransactionIsolation(Connection.TRANSACTION_SERIALIZABLE);
        Assertions.assertEquals(Connection.TRANSACTION_SERIALIZABLE,  conn.getTransactionIsolation());
        closeSQLObjects(conn);
    }

    @Test
    public void testGetWarning() throws SQLException {
        connection.clearWarnings();
        Assertions.assertNull(connection.getWarnings());
    }

    @Test
    public void testCreateStatement() throws SQLException {
        Statement st1 = connection.createStatement(
                ResultSet.TYPE_SCROLL_INSENSITIVE,
                ResultSet.TYPE_FORWARD_ONLY);
        Statement st2 = connection.createStatement(
                ResultSet.TYPE_SCROLL_INSENSITIVE,
                ResultSet.TYPE_FORWARD_ONLY,
                ResultSet.CLOSE_CURSORS_AT_COMMIT
                );

        closeSQLObjects(st1, st2);
    }

    @Test
    void testPrepareStatement() throws SQLException {
        PreparedStatement pst1 = connection.prepareStatement(
                "select 1",
                ResultSet.TYPE_SCROLL_INSENSITIVE,
                ResultSet.TYPE_FORWARD_ONLY
        );

        PreparedStatement pst2 = connection.prepareStatement(
                "select 1",
                ResultSet.TYPE_SCROLL_INSENSITIVE,
                ResultSet.TYPE_FORWARD_ONLY,
                ResultSet.CLOSE_CURSORS_AT_COMMIT
        );

        PreparedStatement pst3 = connection.prepareStatement(
                "select 2",
                Statement.NO_GENERATED_KEYS
        );

        PreparedStatement pst4 = connection.prepareStatement(
                "select 2",
                new int[]{0}
        );

        PreparedStatement pst5 = connection.prepareStatement(
                "select 2",
                new String[]{"column_name"}
        );

        closeSQLObjects(pst1, pst2, pst3, pst4, pst5);
    }

    @Test
    void testCallableStatement() throws SQLException {
        Statement statement = connection.createStatement();
        statement.execute("DROP TABLE citys IF EXISTS");
        statement.executeUpdate("CREATE TABLE citys (id INTEGER, name varchar(20));");
        statement.executeUpdate("INSERT INTO citys VALUES (1, '成都'), (2, '上海'); ");

        CallableStatement cst1 = connection.prepareCall(
                "CREATE PROCEDURE get_citys_name(" +
                        "IN c_id INT, " +
                        "OUT citys_name varchar(20)) " +
                        "READS SQL DATA " +
                        "BEGIN ATOMIC " +
                        "   SELECT citys.name INTO citys_name FROM citys WHERE id = c_id; " +
                        "END;",
                ResultSet.TYPE_SCROLL_INSENSITIVE,
                ResultSet.TYPE_FORWARD_ONLY
        );

        CallableStatement cst2 = connection.prepareCall(
                "CREATE PROCEDURE get_citys_name(" +
                        "IN c_id INT, " +
                        "OUT citys_name varchar(20)) " +
                        "READS SQL DATA " +
                        "BEGIN ATOMIC " +
                        "   SELECT citys.name INTO citys_name FROM citys WHERE id = c_id; " +
                        "END;",
                ResultSet.TYPE_SCROLL_INSENSITIVE,
                ResultSet.TYPE_FORWARD_ONLY,
                ResultSet.CLOSE_CURSORS_AT_COMMIT
        );

        closeSQLObjects(cst1, cst2);
    }

    @Test
    void testCreateArrayOf() throws SQLException {
        Array array = connection.createArrayOf("DOUBLE", new Double[]{0.1, 0.2});
    }

    @Test
    void testCreateStruct() throws SQLException{
        Assertions.assertThrows(SQLException.class, () -> {
            serviceConnection.createStruct( "java.lang.String", new String[]{"hello"});
            RpcRequest request = map.get(null);
            request.setIDToInvoke(IDtoInvoke);
            dispatcher.invokeAsRequest(request, beanCache);
        });
    }

    @Test
    void testAutoCommit() throws SQLException {
        Connection conn = getConnection();
        Assertions.assertTrue(conn.getAutoCommit());
        conn.setAutoCommit(false);
        Assertions.assertFalse(conn.getAutoCommit());
    }

    @Test
    void testGetMetaData() throws SQLException {
        DatabaseMetaData metaData = connection.getMetaData();
    }

    @Test
    void testCreate() throws SQLException{
        Assertions.assertNotNull(connection.createClob());
        Assertions.assertNotNull(connection.createBlob());
        Assertions.assertNotNull(connection.createArrayOf("INTEGER", new Object[]{1}));
        Assertions.assertThrows(SQLFeatureNotSupportedException.class, () -> {
            serviceConnection.createStruct( "java.lang.String", new String[]{"hello"});
            RpcRequest request = map.get(null);
            request.setIDToInvoke(IDtoInvoke);
            dispatcher.invokeAsRequest(request, beanCache);
        });
    }
}
