package com.fanruan;

import org.junit.jupiter.api.*;

import java.sql.*;

/**
 * @author Yichen Dai
 * @date 2022/9/6 9:17
 */
@TestInstance(TestInstance.Lifecycle.PER_CLASS)
public class StatementTest extends BaseJDBCTest{

    Connection connection = null;
    Statement statement = null;

    @BeforeAll
    public void setUp() throws SQLException {
        openSocket();
        connection = getConnection();
        this.statement = connection.createStatement();
    }

    @Test
    public void testGetConnection() throws SQLException{
        Assertions.assertEquals(this.connection, statement.getConnection());
    }

    @Test
    public void testFieldSize() throws SQLException{
        Statement statement = connection.createStatement();
        statement.setMaxFieldSize(1000);
        // HSQLDB always returns zero, meaning there is no limit.
        Assertions.assertEquals(0, statement.getMaxFieldSize());
        closeSQLObjects(statement);
    }

    @Test
    public void testMaxRows() throws SQLException{
        Statement statement = connection.createStatement();
        statement.setMaxRows(1000);
        Assertions.assertEquals(1000, statement.getMaxRows());
        closeSQLObjects(statement);
    }

    @Test
    public void testSet() throws SQLException{
        Statement statement = connection.createStatement();
        statement.setEscapeProcessing(false);

        statement.setCursorName("name");

        statement.setPoolable(false);

        closeSQLObjects(statement);
    }

    @Test
    public void testQueryTimeout() throws SQLException{
        Statement statement = connection.createStatement();
        statement.setQueryTimeout(2000);
        Assertions.assertEquals(2000, statement.getQueryTimeout());
        closeSQLObjects(statement);
    }

    @Test
    public void testWarnings() throws SQLException{
        Statement statement = connection.createStatement();
        statement.clearWarnings();
        Assertions.assertNull(statement.getWarnings());
        closeSQLObjects(statement);
    }

    @Test
    public void testGetResultSet() throws SQLException{
        Statement statement = connection.createStatement();

        String SQL_DROP_IF_EXIST = "DROP TABLE IF EXISTS EMPLOYEE;";

        String SQL_INSERT = "INSERT INTO EMPLOYEE VALUES 'Tim';";

        String SQL_CREATE = "CREATE TABLE EMPLOYEE (NAME varchar(100));";

        statement.execute(SQL_DROP_IF_EXIST);
        statement.execute(SQL_CREATE);
        statement.executeUpdate(SQL_INSERT);

        statement.execute("SELECT * FROM EMPLOYEE");

        ResultSet resultSet = statement.getResultSet();
        Assertions.assertTrue(resultSet.next());
        closeSQLObjects(statement);
    }

    @Test
    public void testDirection() throws SQLException {
        Statement statement = connection.createStatement();
        // Starting with 2.0, HSQLDB accepts any valid value.
        statement.setFetchDirection(ResultSet.FETCH_FORWARD);
        Assertions.assertEquals(ResultSet.FETCH_FORWARD, statement.getFetchDirection());
        closeSQLObjects(statement);
    }

    @Test
    public void testFetchSize() throws SQLException{
        Statement statement = connection.createStatement();
        statement.setFetchSize(0);

        Assertions.assertEquals(0, statement.getFetchSize());
        closeSQLObjects(statement);
    }

    @Test
    public void testWrap() throws SQLException {
        Assertions.assertFalse(statement.isWrapperFor(null));
        Assertions.assertThrows(SQLException.class, () -> statement.unwrap(null));
    }

    @Test
    public void testExecuteBatch() throws SQLException{
        String SQL_DROP_IF_EXIST = "DROP TABLE IF EXISTS EMPLOYEE;";

        String SQL_INSERT = "INSERT INTO EMPLOYEE VALUES 'Tim';";

        String SQL_CREATE = "CREATE TABLE EMPLOYEE (NAME varchar(100));";

        Statement statement = connection.createStatement();
        statement.execute(SQL_DROP_IF_EXIST);
        statement.execute(SQL_CREATE);
        statement.clearBatch();

        statement.addBatch(SQL_INSERT);
        statement.addBatch(SQL_INSERT);
        statement.addBatch(SQL_INSERT);

        int[] count = statement.executeBatch();

        Assertions.assertEquals(3, count.length);
        closeSQLObjects(statement);
    }

    @Test
    public void testGetResultSetInfo() throws SQLException {
        Statement statement = connection.createStatement();

        // HSQLDB supports CONCUR_READ_ONLY and CONCUR_UPDATABLE concurrency.
        Assertions.assertEquals(ResultSet.CONCUR_READ_ONLY, statement.getResultSetConcurrency());

        // HSQLDB 1.7.0 and later versions support TYPE_FORWARD_ONLY and TYPE_SCROLL_INSENSITIVE.
        Assertions.assertEquals(ResultSet.TYPE_FORWARD_ONLY, statement.getResultSetType());

        closeSQLObjects(statement);
    }

    @Test
    public void testGenerateKeys() throws SQLException{
        Statement statement = connection.createStatement();
        Assertions.assertEquals(0, getSizeOfResultSet(statement.getGeneratedKeys()));
        closeSQLObjects(statement);
    }

    @Test
    void testHoldability() throws SQLException{
        Statement statement = connection.createStatement();

        // HSQLDB returns true for both alternatives.
        Assertions.assertEquals(ResultSet.HOLD_CURSORS_OVER_COMMIT, statement.getResultSetHoldability());

        closeSQLObjects(statement);
    }
}
